from functools import wraps
from typing import Union

from .nested_timed import nested_timed
from .simple_timed import timed

_timed_decorators = {}


def create_timed_decorator(
    name: str,
    nested: bool = False,
    enabled: bool = True,
    collect_gc: bool = True,
    disable_gc: bool = False,
    use_seconds: bool = False,
    precision: int = 9,
    show_args: bool = False,
    show_kwargs: bool = False,
    display_level: int = 1,
    sep: str = ", ",
    stdout: bool = True,
    file_path: Union[str, None] = None,
    logger_name: Union[str, None] = None,
    return_time: bool = False,
    out: dict = None,
    use_qualname: bool = False,
):
    """
    Registers a timed decorator with a given name. Once instantiated, the timed decorator can be retrieved with
    :class:`timed_decorator.builder.get_timed_decorator` and used for measuring the runtime of decorated functions if
    enabled.

    Args:
        name (str): The name of the timed decorator which will be instantiated using the provided arguments. Use this
            name for retrieving the timed decorator with :class:`timed_decorator.builder.get_timed_decorator`.
        nested (bool): If `True`, uses the :class:`timed_decorator.nested_timed.nested_timed` as decorator, otherwise
            uses :class:`timed_decorator.simple_timed.timed`. Default: `False`.
        enabled (bool): If `True`, the timed decorator is enabled and used for timing decorated functions. Otherwise,
            functions decorated with `name` will not be timed. Default: `True`.

    See Also:
        :class:`timed_decorator.simple_timed.timed` for the remaining parameters' documentation.

    """
    global _timed_decorators
    if name in _timed_decorators:
        raise KeyError(f"Timed decorator {name} already registered.")

    decorator = nested_timed if nested else timed
    timer = decorator(
        collect_gc=collect_gc,
        disable_gc=disable_gc,
        use_seconds=use_seconds,
        precision=precision,
        show_args=show_args,
        show_kwargs=show_kwargs,
        display_level=display_level,
        sep=sep,
        stdout=stdout,
        file_path=file_path,
        logger_name=logger_name,
        return_time=return_time,
        out=out,
        use_qualname=use_qualname,
    )
    _timed_decorators[name] = (timer, enabled)


def _get_timed_decorator(name: str):
    global _timed_decorators
    if name not in _timed_decorators:
        raise KeyError(
            f"Timed decorator {name} not registered. Please register it first using "
            f"timed_decorator.builder.create_timed_decorator"
        )

    return _timed_decorators[name]


def get_timed_decorator(name: str):
    """
    Wraps the decorated function and lazily measures its elapsed time using the registered timed decorator. The timer
    can be registered after the function definition, but must be registered before the first function call. If the timer
    is disabled, the elapsed time will not be measured.

    Args:
        name (str): The name of the timed decorator registered using
            :class:`timed_decorator.builder.create_timed_decorator`.

    """

    def decorator(fn):
        @wraps(fn)
        def wrap(*args, **kwargs):
            timer, enabled = _get_timed_decorator(name)
            if enabled:
                return timer(fn)(*args, **kwargs)
            return fn(*args, **kwargs)

        return wrap

    return decorator



import logging
from collections.abc import Sequence
from typing import Union

try:
    from numpy import ndarray
except ModuleNotFoundError:

    class ndarray:
        pass


try:
    from pandas import DataFrame
    from pandas import Series
except ModuleNotFoundError:

    class DataFrame:
        pass

    class Series:
        pass


try:
    import torch
    from torch import Tensor
except ModuleNotFoundError:

    class Tensor:
        pass

    torch = None


def nop(*args, **kwargs):
    pass


def build_decorated_fn(fn, decorator, **decorator_kwargs):
    @decorator(**decorator_kwargs)
    def decorated(*args, **kwargs):
        return fn(*args, **kwargs)

    return decorated


class TimeFormatter:
    def __init__(self, use_seconds: bool = False, precision: int = 9):
        self.use_seconds = use_seconds
        self.precision = precision

    def __call__(self, nanoseconds):
        if self.use_seconds:
            return f"{nanoseconds / 1e9:.{self.precision}f}s"
        return f"{nanoseconds}ns"


class Logger:
    def __init__(
        self, stdout: bool, file_path: Union[str, None], logger_name: Union[str, None]
    ):
        self.stdout = stdout
        self.file_path = file_path
        self.logger_name = logger_name

    def __call__(self, string: str):
        if self.stdout:
            print(string)

        if self.file_path is not None:
            with open(self.file_path, "a") as f:
                f.write(string + "\n")

        if self.logger_name is not None:
            logging.getLogger(self.logger_name).info(string)


class InputFormatter:
    def __init__(
        self,
        show_args: bool = False,
        show_kwargs: bool = False,
        display_level: int = 1,
        sep: str = ", ",
    ):
        self.show_args = show_args
        self.show_kwargs = show_kwargs
        self.display_level = display_level
        self.sep = sep

    def __call__(self, fn_name, *args, **kwargs):
        parameters = ""
        if self.show_args:
            parameters = self.format_args(*args)
        if self.show_kwargs:
            kwargs = self.format_kwargs(**kwargs)
            if len(parameters) > 0:
                parameters += self.sep + kwargs
            else:
                parameters = kwargs

        return f"{fn_name}({parameters})"

    def format(self, x) -> str:
        if self.display_level == 0:
            return type(x).__name__

        if self.display_level == 1:
            if isinstance(x, (ndarray, DataFrame, Series)):
                return f"{type(x).__name__}{x.shape}"

            if isinstance(x, Tensor):
                return f'{x.device.type.capitalize()}{type(x).__name__}{str(x.shape).lstrip("torch.Size(").rstrip(")")}'

            if isinstance(x, (str, int, float, bool)):
                return str(x)

            if isinstance(x, Sequence) and len(x) > 0:
                return f"{type(x).__name__}({type(x[0]).__name__})[{len(x)}]"

            if hasattr(x, "__len__"):
                return f"{type(x).__name__}[{len(x)}]"

            return type(x).__name__

        return str(x)

    def format_args(self, *args) -> str:
        return ", ".join([self.format(i) for i in args])

    def format_kwargs(self, **kwargs):
        ret = [(k, self.format(v)) for k, v in kwargs.items()]
        ret = map(str, ret)
        return self.sep.join(ret)


def get_tensor_device(args):
    for i in args:
        if isinstance(i, Tensor) and i.device.type == "cuda":
            return i.device
        if (
            isinstance(i, Sequence)
            and len(i) > 0
            and isinstance(i[0], Tensor)
            and i[0].device.type == "cuda"
        ):
            return i[0].device
    return ""


def synchronize_cuda(*args, **kwargs):
    if torch is None:
        return
    device = get_tensor_device(args)
    if device == "":
        device = get_tensor_device(kwargs.values())
    if device != "":
        torch.cuda.synchronize(device)


def update_timing_dict(out: dict, key: str, elapsed: int, own_time: int):
    counts, elapsed_total, own_time_total = out.get(key, (0, 0, 0))
    out[key] = [counts + 1, elapsed_total + elapsed, own_time_total + own_time]


def get_fn_name(fn):
    if hasattr(fn, "name"):  # torch.jit.ScriptFunction uses name instead of __name__
        return fn.name
    return fn.__name__


def get_fn_qualname(fn):
    if hasattr(
        fn, "qualified_name"
    ):  # torch.jit.ScriptFunction uses qualified_name instead of __qualname__
        return fn.qualified_name
    return fn.__qualname__



import logging
import unittest
from io import StringIO

import torch
from torch import Tensor

from timed_decorator.simple_timed import timed


def exp_plus_mean(x: Tensor) -> Tensor:
    return x.exp() + x.mean()


class MyTestCase(unittest.TestCase):
    def setUp(self):
        self.logger_name = "TEST_LOGGER"
        self.log_stream = StringIO()
        log_handler = logging.StreamHandler(self.log_stream)
        logging.root.setLevel(logging.NOTSET)
        logging.getLogger(self.logger_name).addHandler(log_handler)
        self.fn = exp_plus_mean

    def do_tensor_op(self, x: Tensor):
        timed(logger_name=self.logger_name, stdout=False, show_args=True)(self.fn)(x)

    def test_tensor_name(self):
        cpu_tensor = torch.rand((10, 50))
        self.do_tensor_op(cpu_tensor)
        logged = self.log_stream.getvalue().split("\n")[0]
        self.assertIn("CpuTensor[10, 50]", logged)

    def test_tensor_name_cuda(self):
        if torch.cuda.is_available():
            cuda_tensor = torch.rand((10, 50), device="cuda:0")
            self.fn(cuda_tensor)
            logged = self.log_stream.getvalue().split("\n")[0]
            self.assertIn("CudaTensor[10, 50]", logged)

    def test_traced_fn(self):
        self.fn = torch.jit.trace(exp_plus_mean, example_inputs=torch.rand((10, 50)))

        cpu_tensor = torch.rand((10, 50))
        self.do_tensor_op(cpu_tensor)
        logged = self.log_stream.getvalue().split("\n")[0]
        self.assertIn("CpuTensor[10, 50]", logged)

    def test_scripted_fn(self):
        self.fn = torch.jit.script(exp_plus_mean)

        cpu_tensor = torch.rand((10, 50))
        self.do_tensor_op(cpu_tensor)
        logged = self.log_stream.getvalue().split("\n")[0]
        self.assertIn("CpuTensor[10, 50]", logged)


if __name__ == "__main__":
    unittest.main()



import gc
from functools import wraps
from gc import collect
from time import perf_counter_ns
from typing import Union

from .utils import (
    nop,
    TimeFormatter,
    InputFormatter,
    synchronize_cuda,
    Logger,
    update_timing_dict,
    get_fn_name,
    get_fn_qualname,
)


def timed(
    collect_gc: bool = True,
    disable_gc: bool = False,
    use_seconds: bool = False,
    precision: int = 9,
    show_args: bool = False,
    show_kwargs: bool = False,
    display_level: int = 1,
    sep: str = ", ",
    stdout: bool = True,
    file_path: Union[str, None] = None,
    logger_name: Union[str, None] = None,
    return_time: bool = False,
    out: dict = None,
    use_qualname: bool = False,
):
    """
    A simple timing decorator that measures the time elapsed during the function call and prints it.
    It uses perf_counter_ns for measuring which includes time elapsed during sleep and is system-wide.

    Args:
        collect_gc (bool): If `True`, runs a full garbage collection before timing the wrapped function. Default:
            `True`.
        disable_gc (bool): If `True`, disabled garbage collection during function execution. Default: `False`.
        use_seconds (bool): If `True`, displays the elapsed time in seconds. Default: `False`.
        precision (int): Used in conjunction with `use_seconds`, represents the decimal points used for printing
            seconds. Default: `9`.
        show_args (bool): If `True`, displays the function arguments according to `display_level`. Useful when timing
            function calls with arguments of different magnitude. Default: `False`.
        show_kwargs (bool): If `True`, displays the keyword arguments according to `display_level`. Default: `False`.
        display_level (int): The level of verbosity used when printing function arguments ad keyword arguments. If `0`,
            prints the type of the parameters. If `1`, prints values for all primitive types, shapes for arrays,
            tensors, dataframes and length for sequences. Otherwise, prints values for all parameters. Default: `1`.
        sep (str): The separator used when printing function arguments and keyword arguments. Default: `', '`.
        stdout (bool): If `True`, writes the elapsed time to stdout. Default: `True`.
        file_path (str): If not `None`, writes the measurement at the end of the given file path. For thread safe
            file writing configure use `logger_name` instead. Default: `None`.
        logger_name (str): If not `None`, uses the given logger to print the measurement. Can't be used in conjunction
            with `file_path`. Default: `None`.
        return_time (bool): If `True`, returns the elapsed time in addition to the wrapped function's return value.
            Default: `False`.
        out (dict): If not `None`, stores the elapsed time in nanoseconds in the given dict using the fully qualified
            function name as key, in the following format: (function call counts, total elapsed time, total "own time").
            If the key already exists, updates the existing value. The elapsed time is equal to "own time" for the simple
            timed decorator. For the nested time decorator, the elapsed time is different from "own time" only when
            another function decorated with a nested timer is called during the execution of the current function.
            Default: `None`.
        use_qualname (bool): If `True`, uses the qualified name of the function when logging the elapsed time. Default:
            `False`.
    """
    gc_collect = collect if collect_gc else nop
    time_formatter = TimeFormatter(use_seconds, precision)
    input_formatter = InputFormatter(show_args, show_kwargs, display_level, sep)
    logger = Logger(stdout, file_path, logger_name)
    update_dict = update_timing_dict if out is not None else nop

    def decorator(fn):
        @wraps(fn)
        def wrap(*args, **kwargs):
            gc_collect()
            if disable_gc:
                gc.disable()

            try:
                start = perf_counter_ns()
                ret = fn(*args, **kwargs)
                synchronize_cuda(*args, **kwargs)
                end = perf_counter_ns()
            finally:
                if disable_gc:
                    gc.enable()

            elapsed = end - start
            fn_name = get_fn_qualname(fn) if use_qualname else get_fn_name(fn)
            update_dict(out, get_fn_qualname(fn), elapsed, elapsed)
            logger(
                f"{input_formatter(fn_name, *args, **kwargs)} -> total time: {time_formatter(elapsed)}"
            )
            if return_time:
                return ret, elapsed
            return ret

        return wrap

    return decorator



import logging
import os
import unittest
from io import StringIO
from time import sleep

from tests.functions import fibonacci, recursive_fibonacci
from timed_decorator.builder import create_timed_decorator, get_timed_decorator
from timed_decorator.nested_timed import nested_timed
from timed_decorator.simple_timed import timed
from timed_decorator.utils import build_decorated_fn


class UsageTest(unittest.TestCase):
    @staticmethod
    def test_simple_timed():
        fn = build_decorated_fn(fibonacci, timed)
        fn(1000)

        fn = build_decorated_fn(fibonacci, timed, collect_gc=False)
        fn(1000)

        fn = build_decorated_fn(recursive_fibonacci, timed)
        fn(30)

        fn = build_decorated_fn(recursive_fibonacci, timed, collect_gc=False)
        fn(30)

        fn = build_decorated_fn(
            recursive_fibonacci, timed, collect_gc=False, use_seconds=True
        )
        fn(30)

        fn = build_decorated_fn(
            recursive_fibonacci, timed, collect_gc=False, use_seconds=True, precision=5
        )
        fn(30)

        fn = build_decorated_fn(fibonacci, timed, show_args=True, display_level=0)
        fn(1000)

        fn = build_decorated_fn(fibonacci, timed, show_args=True, display_level=2)
        fn(1000)

        fn = build_decorated_fn(fibonacci, timed, show_args=True, disable_gc=True)
        fn(1000)

    @staticmethod
    def test_nested_timed():
        @nested_timed(collect_gc=False, use_seconds=True, precision=3)
        def nested_fn():
            @nested_timed(collect_gc=False, use_seconds=True, precision=3)
            def sleeping_fn(x):
                sleep(x)

            @nested_timed(collect_gc=False, use_seconds=True, precision=3)
            def other_fn():
                sleep(0.1)
                sleeping_fn(0.1)

            sleep(0.1)
            sleeping_fn(0.1)
            other_fn()
            sleeping_fn(0.1)

        nested_fn()

    def test_file_usage(self):
        filename = "file.txt"

        @timed(file_path=filename, stdout=False)
        def fn():
            sleep(0.1)

        try:
            fn()
            fn()
            with open(filename, "r") as f:
                lines = f.readlines()
                self.assertEqual(len(lines), 2)
                for line in lines:
                    self.assertEqual(line[:2], fn.__name__)
        finally:
            if os.path.isfile(filename):
                os.remove(filename)

    def test_logger_usage(self):
        logger_name = "TEST_LOGGER"
        log_stream = StringIO()
        log_handler = logging.StreamHandler(log_stream)
        logging.root.setLevel(logging.NOTSET)
        logging.getLogger(logger_name).addHandler(log_handler)

        @timed(logger_name=logger_name, stdout=False)
        def fn_1():
            sleep(0.1)

        fn_1()

        @timed(logger_name=logger_name, stdout=False, use_qualname=True)
        def fn_2():
            sleep(0.1)

        fn_2()

        logged = log_stream.getvalue().split("\n")[:-1]
        self.assertEqual(len(logged), 2)
        self.assertIn(fn_1.__name__, logged[0])
        self.assertNotIn(fn_1.__qualname__, logged[0])
        self.assertIn(fn_2.__name__, logged[1])
        self.assertIn(fn_2.__qualname__, logged[1])

    def test_dict_output(self):
        out = {}

        @timed(out=out, stdout=False)
        def fn():
            sleep(0.1)

        fn()

        counts, elapsed, own_time = out[fn.__qualname__]
        self.assertEqual(counts, 1)
        self.assertIsInstance(elapsed, int)
        self.assertGreater(elapsed, 1e8)
        self.assertEqual(elapsed, own_time)

        fn()
        counts, elapsed, own_time = out[fn.__qualname__]
        self.assertEqual(counts, 2)
        self.assertIsInstance(elapsed, int)
        self.assertGreater(elapsed, 2e8)
        self.assertEqual(elapsed, own_time)

    def test_dict_output_with_nested_timed(self):
        out = {}

        @nested_timed(out=out, stdout=False)
        def fn1():
            @nested_timed(out=out, stdout=False)
            def fn2():
                sleep(0.1)

            fn2()

        fn1()

        counts_1, elapsed_1, own_time_1 = out[fn1.__qualname__]
        counts_2, elapsed_2, own_time_2 = out[fn1.__qualname__ + ".<locals>.fn2"]

        self.assertTrue(counts_1 == counts_2 == 1)
        self.assertEqual(own_time_2, elapsed_2)
        self.assertGreater(elapsed_1, own_time_1)

    def test_qualname(self):
        out = {}

        class ClassA:
            @timed(out=out)
            def wait(self, x):
                sleep(x)

        ClassA().wait(0.1)
        key = next(iter(out.keys()))
        self.assertIn(ClassA.__name__, key)

    def test_return_time(self):
        seconds = 0.1

        @timed(return_time=True, stdout=False)
        def fn(x):
            sleep(x)

        _, elapsed = fn(seconds)

        self.assertIsInstance(elapsed, int)
        self.assertGreater(elapsed / 1e9, seconds)

    def test_create_timed_decorator(self):
        create_timed_decorator("same name")
        self.assertRaises(KeyError, create_timed_decorator, "same name")
        create_timed_decorator("other name")

        get_timed_decorator("same name")(fibonacci)(10000)
        get_timed_decorator("other name")(fibonacci)(10000)
        fn = get_timed_decorator("no name")(
            fibonacci
        )  # Does not raise error if not called
        self.assertRaises(KeyError, fn, 10000)

        # Defer instantiation
        fn = get_timed_decorator("lazy name")(fibonacci)
        create_timed_decorator("lazy name")
        fn(10000)

    def test_create_disabled_timed_decorator(self):
        out = {}
        create_timed_decorator("disabled", enabled=False, out=out)
        get_timed_decorator("disabled")(fibonacci)(10000)

        create_timed_decorator("enabled", enabled=True, out=out)
        get_timed_decorator("enabled")(recursive_fibonacci)(30)

        self.assertNotIn(fibonacci.__qualname__, out)
        self.assertIn(recursive_fibonacci.__qualname__, out)


if __name__ == "__main__":
    unittest.main()



import gc
from functools import wraps
from gc import collect
from time import perf_counter_ns
from typing import Union

from .utils import (
    nop,
    TimeFormatter,
    InputFormatter,
    synchronize_cuda,
    Logger,
    update_timing_dict,
    get_fn_name,
    get_fn_qualname,
)

nested_level = 0
nested_times = dict()


def nested_timed(
    collect_gc: bool = True,
    disable_gc: bool = False,
    use_seconds: bool = False,
    precision: int = 9,
    show_args: bool = False,
    show_kwargs: bool = False,
    display_level: int = 1,
    sep: str = ", ",
    stdout: bool = True,
    file_path: Union[str, None] = None,
    logger_name: Union[str, None] = None,
    return_time: bool = False,
    out: dict = None,
    use_qualname: bool = False,
):
    """
    A nested timing decorator that measures the time elapsed during the function call and accounts for other decorators
    further in the call stack.

    See Also:
        :class:`timed_decorator.simple_timed.timed` for parameter documentation.
    """
    gc_collect = collect if collect_gc else nop
    time_formatter = TimeFormatter(use_seconds, precision)
    input_formatter = InputFormatter(show_args, show_kwargs, display_level, sep)
    logger = Logger(stdout, file_path, logger_name)
    update_dict = update_timing_dict if out is not None else nop

    def decorator(fn):
        @wraps(fn)
        def wrap(*args, **kwargs):
            global nested_level, nested_times
            nested_level += 1

            gc_collect()
            if disable_gc:
                gc.disable()

            try:
                start = perf_counter_ns()
                ret = fn(*args, **kwargs)
                synchronize_cuda(*args, **kwargs)
                end = perf_counter_ns()
            except Exception as e:
                nested_level -= 1
                raise e
            finally:
                if disable_gc:
                    gc.enable()

            elapsed = end - start

            children_time = 0
            if nested_level in nested_times:
                children_time = sum(nested_times[nested_level])
                del nested_times[nested_level]

            own_time = elapsed - children_time
            nested_level -= 1

            if nested_level != 0:
                if nested_level not in nested_times:
                    nested_times[nested_level] = []
                nested_times[nested_level].append(elapsed)

            fn_name = get_fn_qualname(fn) if use_qualname else get_fn_name(fn)
            update_dict(out, get_fn_qualname(fn), elapsed, own_time)
            logger(
                "\t" * nested_level + f"{input_formatter(fn_name, *args, **kwargs)} "
                f"-> total time: {time_formatter(elapsed)}, "
                f"own time: {time_formatter(own_time)}"
            )
            if return_time:
                return ret, elapsed
            return ret

        return wrap

    return decorator
